from typing import Any

from django_accounts_api.forms.user_create import UnusablePasswordUserCreationForm
from django_accounts_api.views.users import _user_has_user_perm

from django.http import HttpRequest, HttpResponse, HttpResponseForbidden, JsonResponse
from django.views.generic.edit import FormView

from .base import AcceptJsonMixin
from .UserReset import ResetPasswordMixin

class UserCreate(AcceptJsonMixin, ResetPasswordMixin, FormView):
    """ Create a user from the basic name and email information
    Do not set the password, but generate a link to set it
    """
    form_class = UnusablePasswordUserCreationForm
    template_name = 'django_accounts_api/form.html'

    def dispatch(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        if not request.user.is_authenticated:
            return HttpResponse(status=401)
        if not _user_has_user_perm(request.user, 'add'):
            return HttpResponseForbidden()
        return super().dispatch(request, *args, **kwargs)

    def render_to_response(self, context: dict[str, Any], **response_kwargs: Any) -> HttpResponse:
        if self.json_response_requested():
            return JsonResponse(
                [
                    {
                        'type': 'text'
                    }
                ],
                safe=False
            )
        else:
            return super().render_to_response(context, **response_kwargs)

    def form_invalid(self, form: Any) -> HttpResponse:
        if self.json_response_requested():
            response = JsonResponse(
                form.errors
            )
            response.status_code = 400
        else:
            response = super().form_invalid(form)
            response.status_code = 400
        return response

    
    def form_valid(self, form: Any) -> HttpResponse:
        self.user = form.save()
        self.set_reset_link(self.user)
        return HttpResponse(self.reset_link, status=200)


import re
from typing import Any, Dict

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser
from django.core import serializers
from django.forms.models import BaseModelForm
from django.http import HttpRequest, HttpResponse, HttpResponseForbidden, JsonResponse
from django.utils.decorators import method_decorator
from django.views.decorators.cache import never_cache
from django.views.generic.edit import UpdateView, DeleteView
from django.views.generic.list import BaseListView


User = get_user_model()

# Setting to control the fields returned by the user api
USER_FIELD_SETTTING = 'ACCOUNTS_API_USER_FIELDS'
# DEfault fields returned byt he user aPI if no setting provided
DEFAULT_USER_FIELDS = [
    'username',
    'first_name',
    'last_name',
    'email',
    'last_login',
    'is_active',
]

def _is_user_perm(perm_codename: str):
    ''' returns true if the permission codename relates to the configured user model'''
    match =  re.match(
        f'^{User._meta.app_label}\.[\w]+_{User._meta.model_name}',
        perm_codename
    )
    return match

def _user_has_any_user_perm(user):
    ''' returns true if the user has any permission related to the configured user model'''
    perms = user.get_all_permissions()
    return any(
        filter(
            lambda perm: _is_user_perm(perm),
            perms
        )
    )

def _user_has_user_perm(user: AbstractUser, perm: str):
    ''' returns true if the user has the specified permission on the configured user model'''
    return user.has_perm(f'{User._meta.app_label}.{perm}_{User._meta.model_name}')

class JsonListView(BaseListView):
    ''' Extends the django base list view to return json serialized data
    '''
    def render_to_response(self, context, **response_kwargs):
        queryset = self.get_queryset()
        response = HttpResponse(
           serializers.serialize("json", queryset, fields=self.get_fields())
        )
        response.headers['Content-Type'] = 'application/json'
        return response

@method_decorator(never_cache, name='dispatch')
class APIUsersView(JsonListView):
    ''' Returns a json serialized list of user models
    
    
    '''
    model = User
    fields_setting_name = None

    def dispatch(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        if not request.user.is_authenticated:
            return HttpResponse(status=401)
        if not _user_has_any_user_perm(request.user):
            return HttpResponseForbidden()
        return super().dispatch(request, *args, **kwargs)

    
    def get_fields(self):
        return getattr(settings, USER_FIELD_SETTTING, DEFAULT_USER_FIELDS)

class UserUpdate(UpdateView):
    model = User
    fields = ['first_name', 'last_name', 'email', 'is_active']
    template_name = 'django_accounts_api/schemas/user_update.json'
    def dispatch(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        if not request.user.is_authenticated:
            return HttpResponse(status=401)
        if not _user_has_user_perm(request.user, 'change'):
            return HttpResponseForbidden()
        return super().dispatch(request, *args, **kwargs)

    def form_invalid(self, form: BaseModelForm) -> HttpResponse:
        response = JsonResponse(form.errors)
        response.status_code = 400
        return response

    def form_valid(self, form: BaseModelForm) -> HttpResponse:
        form.save()
        return HttpResponse(status=200)

    def render_to_response(self, context: Dict[str, Any], **response_kwargs: Any) -> HttpResponse:
        response = super().render_to_response(context, **response_kwargs)
        response['Content-Type'] = 'application/json'
        return response


class UserDelete(DeleteView):
    model = User
    def dispatch(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        if not request.user.is_authenticated:
            return HttpResponse(status=401)
        if not _user_has_user_perm(request.user, 'delete'):
            return HttpResponseForbidden()
        return super().dispatch(request, *args, **kwargs)

    def form_valid(self, form: BaseModelForm) -> HttpResponse:
        self.object.delete()
        return HttpResponse(status=200)

    def render_to_response(self, context: Dict[str, Any], **response_kwargs: Any) -> HttpResponse:
        return HttpResponse(status=200)







from django.test import TestCase, override_settings
from django.urls import reverse
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode
from django.contrib.auth.tokens import default_token_generator
from django.contrib.auth.views import (
    INTERNAL_RESET_SESSION_TOKEN
)

User = get_user_model()
test_user_email = "admin@admin.com"


class PasswordResetTestCase(TestCase):
    """Basic tests of the partial HTML password reset view"""

    def setUp(self) -> None:
        """Get the url and create a user"""
        self.user = User.objects.create_user(
            "test", password="test", email=test_user_email
        )
        self.valid_token = default_token_generator.make_token(self.user)
        self.reset_url_token = 'set-password'
        self.email_url_bad_user = reverse(
            "django_accounts_api:password_reset_confirm",
            kwargs=dict(
                uidb64=urlsafe_base64_encode(force_bytes('nonexistant')),
                token=self.valid_token
            )
        )
        self.email_url = reverse(
            "django_accounts_api:password_reset_confirm",
            kwargs=dict(
                uidb64=urlsafe_base64_encode(force_bytes(self.user.pk)),
                token=default_token_generator.make_token(self.user)
            )
        )
        self.email_invalid_token_url = reverse(
            "django_accounts_api:password_reset_confirm",
            kwargs=dict(
                uidb64=urlsafe_base64_encode(force_bytes(self.user.pk)),
                token="invalid_token"
            )
        )
        self.reset_url = reverse(
            "django_accounts_api:password_reset_confirm",
            kwargs=dict(
                uidb64=urlsafe_base64_encode(force_bytes(self.user.pk)),
                token=self.reset_url_token
            )
        )
        return super().setUp()

    def set_token_in_session(self):
        # Set the session token - must set as variable and call save!
        session = self.client.session
        session[INTERNAL_RESET_SESSION_TOKEN] = self.valid_token
        session.save()

    @override_settings(
        ROOT_URLCONF='testproject.urls_reset_misconfigured'
    )
    def test_password_reset_confirm_misconfigured(self):
        ''' Should return 400 invalid token '''
        with self.assertRaises(ImproperlyConfigured):
            self.client.get(
                '/api/bad'
            )

    def test_password_reset_confirm_bad_user(self):
        ''' Should redirect to the frontend invalid reset code page '''
        response = self.client.get(
            self.email_url_bad_user
        )
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], '/reset-password/invalid/')

    def test_password_reset_confirm_invalid_code_get(self):
        ''' Should redirect to the frontend invalid reset code page '''
        response = self.client.get(
            self.email_invalid_token_url
        )
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], '/reset-password/invalid/')

    def test_password_reset_confirm_valid_get(self):
        ''' Should redirect to the frontend reset password page, placing the token in the sesssion'''
        response = self.client.get(
            self.email_url,
        )
        self.assertEqual(response.status_code, 302)
        location = response.headers["Location"]
        self.assertTrue(location.startswith('/reset'))
        session_token = self.client.session.get(INTERNAL_RESET_SESSION_TOKEN)
        self.assertEqual(session_token, self.valid_token)

    def test_password_reset_confirm_post_success(self):
        data = {
            'new_password1': 'newpass123',
            'new_password2': 'newpass123',
        }

        self.set_token_in_session()

        session = self.client.session
        session[INTERNAL_RESET_SESSION_TOKEN] = self.valid_token
        session.save()

        response = self.client.post(
            self.reset_url, data,
        )
        self.assertEqual(response.status_code, 200)
        self.user.refresh_from_db()
        self.assertTrue(self.user.check_password('newpass123'))

    @override_settings(
        CHANGE_PASSWORD_POST_LOGIN=True
    )
    def test_password_reset_confirm_post_success_login(self):
        data = {
            'new_password1': 'newpass123',
            'new_password2': 'newpass123',
        }

        self.set_token_in_session()

        response = self.client.post(
            self.reset_url, data,
        )
        self.assertEqual(response.status_code, 200)
        self.user.refresh_from_db()
        self.assertTrue(self.user.check_password('newpass123'))
        self.assertIn('_auth_user_id', self.client.session)

    def test_password_reset_confirm_post_form_invalid(self):
        data = {
            'new_password1': 'newpass123',
            'new_password2': 'not_good',
        }

        self.set_token_in_session()

        response = self.client.post(
            self.reset_url, data,
        )
        self.assertEqual(response.status_code, 400)
        assert "form" in response.context
        self.assertEqual(len(response.context['form'].errors), 1)

    def test_password_reset_confirm_post_form_invalid_json(self):
        data = {
            'new_password1': 'newpass123',
            'new_password2': 'not_good',
        }

        self.set_token_in_session()

        response = self.client.post(
            self.reset_url, data,
            HTTP_ACCEPT='application/json'
        )
        self.assertEqual(response.status_code, 400)
        responseContent = response.json()
        self.assertEqual(len(responseContent['errors']), 1)

    '''
    def test_password_reset_confirm_post_no_token(self):
        # prepare the data with invalid token
        data = {
            'new_password1': 'newpass123',
            'new_password2': 'newpass123',
        }
        response = self.client.post(
            self.reset_url, data,
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.status_text, 'invalid request')
    '''



from http import HTTPStatus
from typing import Any
import warnings

from django.contrib.auth.models import User as DjangoUser
from django.contrib.auth.views import LoginView, LogoutView
from django.http import HttpRequest, HttpResponse, JsonResponse
from django.utils.decorators import method_decorator
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import ensure_csrf_cookie
from django.views.decorators.debug import sensitive_post_parameters

from .base import AcceptJsonMixin, user_details


@never_cache
def login_check(request) -> HttpResponse:
    """
    200 and details if logged in, 401 if not

    User details are basic, but can be expanded by providing a dotted import path to a 
    function returning a json serializable dict from the user parameter
    """
    warnings.warn("This endpoint is deprecated", DeprecationWarning)
    user: DjangoUser = request.user
    if (user.is_authenticated):
        return JsonResponse(user_details(user))
    else:
        return HttpResponse(status=401)


@method_decorator(ensure_csrf_cookie, name='get')
@method_decorator(sensitive_post_parameters(), name='dispatch')
class Login(AcceptJsonMixin, LoginView):
    '''
    Override the Django login view to be API friendly for json or partial html

    GET:

    - default: renders a partial login form
    - json requested: json form schema

    POST success: 200 logs the user in
     
    POST failure: 400

    - default: renders a password form with errors
    - json requested: returns password form errors

    '''
    template_name = "django_accounts_api/login.html"

    def form_valid(self, form):
        """Override redirect behavior to return JSON user details"""
        _repressed_redirect = super().form_valid(form)  # noqa: F841
        return JsonResponse(
            user_details(self.request.user),
            status=201
        )

    def form_invalid(self, form):
        """Override redirect behavior if json is requested return json errors"""
        if self.json_response_requested():
            return JsonResponse(dict(errors=form.errors), status=400)
        else:
            return super().form_invalid(form)

    def get(self, request: HttpRequest, *args: str, **kwargs: Any) -> HttpResponse:
        """Override the get behavior if json requested return user details"""
        if self.json_response_requested():
            if (request.user.is_authenticated):
                return JsonResponse(user_details(request.user))
            else:
                return JsonResponse({}, status=HTTPStatus.NO_CONTENT)
        else:
            return super().get(request, *args, **kwargs)


class Logout(LogoutView):
    ''' Override the Django logout view to NOT redirect on successful login
    
    POST - logs out, returns 200
    '''

    def post(self, request, *args, **kwargs):
        _repressed_redirect_or_render = super().post(request, *args, **kwargs)  # noqa: F841
        return HttpResponse(
            status=200
        )



from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission, AbstractBaseUser
from django.contrib.contenttypes.models import ContentType
from django.core import mail
from django.test import TestCase
from django.urls import reverse

User = get_user_model()

test_user_email = "admin@admin.com"
test_user_email2 = "admin2@admin.com"

def add_user_permission(user: AbstractBaseUser, permission: str):
    user.user_permissions.add(
        Permission.objects.get(
            content_type=ContentType.objects.get_for_model(User),
            codename=f"{permission}_{user._meta.model_name}"
        )
    )


class UsersTestCase(TestCase):
    '''Basic tests of the users view'''

    def setUp(self) -> None:
        '''Get the url and create a user'''
        self.url = reverse("django_accounts_api:users")
        self.user = User.objects.create_user("test", password="test", email=test_user_email)
        return super().setUp()

    def test_users_get_no_auth(self):
        '''An unauthed get should get a 401'''
        response = self.client.get(self.url)
        assert response.status_code == 401

    def test_users_get_no_perms(self):
        '''An authed user with no user permissions gshould get a 403 '''
        self.client.force_login(self.user)
        response = self.client.get(
            self.url,
        )
        assert response.status_code == 403

    def test_users_get_with_perms(self):
        '''An authed user with a user permissions gshould get a 200 and a json response'''
        add_user_permission(self.user, 'view')
        self.client.force_login(self.user)
        response = self.client.get(
            self.url,
        )
        assert response.status_code == 200
        json = response.json()
        self.assertEqual(len(json), 1)


class UserUpdateTestCase(TestCase):
    '''Basic tests of the user update view'''

    def setUp(self) -> None:
        '''Get the url and create a user'''
        self.user = User.objects.create_user("test", password="test", email=test_user_email)
        self.user2 = User.objects.create_user("test2", password="test2", email=test_user_email2)
        self.url = reverse("django_accounts_api:user_update", kwargs={"pk": self.user2.pk})
        return super().setUp()

    def test_user_update_get_no_auth(self):
        '''An unauthed get should receive a 401'''
        response = self.client.get(self.url)
        assert response.status_code == 401

    def test_user_update_get_no_perms(self):
        '''An authed user with no user permissions should receive a 403 '''
        self.client.force_login(self.user)
        response = self.client.get(
            self.url,
        )
        assert response.status_code == 403

    def test_user_update_post_no_auth(self):
        '''An unauthed post should receive a 401'''
        response = self.client.post(self.url)
        assert response.status_code == 401

    def test_user_update_post_no_perms(self):
        '''An authed user with no user permissions should receive a 403 '''
        self.client.force_login(self.user)
        response = self.client.post(
            self.url,
        )
        assert response.status_code == 403

    def test_user_update_get_with_perms(self):
        '''An authed user with a user permissions should get a 200 and a json response'''
        add_user_permission(self.user, 'change')
        self.client.force_login(self.user)
        response = self.client.get(
            self.url,
        )
        assert response.status_code == 200
        json = response.json()
        self.assertEqual(len(json), 3)

    def test_user_update_post_invalid(self):
        '''Invalid details posted should respond with a 400 '''
        add_user_permission(self.user, 'change')
        self.client.force_login(self.user)
        response = self.client.post(
            self.url,
            {
                'first_name': 'q' * 200
            }
        )
        assert response.status_code == 400
        json = response.json()
        self.assertEqual(len(json), 1)

    def test_user_update_post_valid(self):
        '''valid details posted should respond with a 200 '''
        add_user_permission(self.user, 'change')
        self.client.force_login(self.user)
        response = self.client.post(
            self.url,
            {
                'first_name': 'changed1',
                'last_name': 'changed2',
                'email': 'admin3@admin.com',
            }
        )
        assert response.status_code == 200
        self.user2.refresh_from_db()
        self.assertEqual(self.user2.first_name, 'changed1')
        self.assertEqual(self.user2.last_name, 'changed2')
        self.assertEqual(self.user2.email, 'admin3@admin.com')


class UserDeleteTestCase(TestCase):
    '''Basic tests of the user delete view'''

    def setUp(self) -> None:
        '''Get the url and create a user'''
        self.user = User.objects.create_user("test", password="test", email=test_user_email)
        self.user2 = User.objects.create_user("test2", password="test2", email=test_user_email2)
        self.url = reverse("django_accounts_api:user_delete", kwargs={"pk": self.user2.pk})
        return super().setUp()

    def test_user_delete_get_no_auth(self):
        '''An unauthed get should receive a 401'''
        response = self.client.get(self.url)
        assert response.status_code == 401

    def test_user_delete_get_no_perms(self):
        '''An authed user with no user permissions should receive a 403 '''
        self.client.force_login(self.user)
        response = self.client.get(
            self.url,
        )
        assert response.status_code == 403

    def test_user_delete_post_no_auth(self):
        '''An unauthed post should receive a 401'''
        response = self.client.post(self.url)
        assert response.status_code == 401

    def test_user_delete_post_no_perms(self):
        '''An authed user with no user permissions should receive a 403 '''
        self.client.force_login(self.user)
        response = self.client.post(
            self.url,
        )
        assert response.status_code == 403

    def test_user_delete_get_with_perms(self):
        '''An authed user with a user permissions should get a 200 and a json response'''
        add_user_permission(self.user, 'delete')
        self.client.force_login(self.user)
        response = self.client.get(
            self.url,
        )
        assert response.status_code == 200

    def test_user_delete_post_valid(self):
        '''valid details posted should respond with a 200 and delete the user'''
        add_user_permission(self.user, 'delete')
        self.client.force_login(self.user)
        response = self.client.post(
            self.url
        )
        assert response.status_code == 200
        with self.assertRaises(User.DoesNotExist):
            self.user2.refresh_from_db()


class UserCreateTestCase(TestCase):
    '''Basic tests of the user create view'''

    def setUp(self) -> None:
        '''Get the url and create a user'''
        self.user = User.objects.create_user("test", password="test", email=test_user_email)
        self.url = reverse("django_accounts_api:user_create")
        return super().setUp()

    def test_user_create_get_no_auth(self):
        '''An unauthed get should receive a 401'''
        response = self.client.get(self.url)
        assert response.status_code == 401

    def test_user_create_get_no_perms(self):
        '''An authed user with no user permissions should receive a 403 '''
        self.client.force_login(self.user)
        response = self.client.get(
            self.url,
        )
        assert response.status_code == 403

    def test_user_create_post_no_auth(self):
        '''An unauthed post should receive a 401'''
        response = self.client.post(self.url)
        assert response.status_code == 401

    def test_user_create_post_no_perms(self):
        '''An authed user with no user permissions should receive a 403 '''
        self.client.force_login(self.user)
        response = self.client.post(
            self.url,
        )
        assert response.status_code == 403

    def test_user_create_get_with_perms(self):
        '''An authed user with a user permissions should get a 200 and a json response'''
        add_user_permission(self.user, 'add')
        self.client.force_login(self.user)
        response = self.client.get(
            self.url,
        )
        assert response.status_code == 200
        assert 'form' in response.context

    def test_user_create_get_with_perms_json(self):
        '''An authed user with a user permissions should get a 200 and a json response'''
        add_user_permission(self.user, 'add')
        self.client.force_login(self.user)
        response = self.client.get(
            self.url,
            HTTP_ACCEPT='application/json'
        )
        assert response.status_code == 200
        json = response.json()

    def test_user_create_post_invalid(self):
        '''valid details posted should respond with a 400 and form eith errors'''
        add_user_permission(self.user, 'add')
        self.client.force_login(self.user)
        response = self.client.post(
            self.url,
            dict(
                username='new user',
            )
        )
        assert response.status_code == 400
    
    def test_user_create_post_json_invalid(self):
        '''valid details posted should respond with a 400 and form eith errors'''
        add_user_permission(self.user, 'add')
        self.client.force_login(self.user)
        response = self.client.post(
            self.url,
            dict(
                username='new user',
            ),
            HTTP_ACCEPT='application/json'
        )
        assert response.status_code == 400
    
    def test_user_create_post_valid(self):
        '''valid details posted should respond with a 200 and create the user'''
        add_user_permission(self.user, 'add')
        self.client.force_login(self.user)
        response = self.client.post(
            self.url,
            dict(
                username='new_user',
                first_name='New',
                last_name='User',
                email='new@user.com'
            )
        )
        assert response.status_code == 200
        new_user: AbstractBaseUser = User.objects.get(username='new_user')
        self.assertEqual(new_user.first_name, 'New')
        self.assertEqual(new_user.last_name, 'User')
        self.assertEqual(new_user.email, 'new@user.com')
        self.assertFalse(new_user.has_usable_password())
        self.assertEqual(len(mail.outbox), 1)

class UserResetPasswordTestCase(TestCase):
    '''Basic tests of the user reset password view'''

    def setUp(self) -> None:
        '''Get the url and create a user'''
        self.user = User.objects.create_user("test", password="test", email=test_user_email)
        self.not_found_url = reverse("django_accounts_api:user_reset", kwargs=dict(pk=999999))
        self.url = reverse("django_accounts_api:user_reset", kwargs=dict(pk=self.user.id))
        return super().setUp()
    

    def test_user_reset_post_no_auth(self):
        '''An unauthed post should receive a 401'''
        response = self.client.post(self.url)
        assert response.status_code == 401

    def test_user_create_post_no_perms(self):
        '''An authed user with no user permissions should receive a 403 '''
        self.client.force_login(self.user)
        response = self.client.post(
            self.url,
        )
        assert response.status_code == 403

    def test_user_reset_get_not_supported(self):
        '''An unauthed post should receive a method not supported'''
        add_user_permission(self.user, 'change')
        self.client.force_login(self.user)
        response = self.client.get(self.url)
        assert response.status_code == 405
    
    def test_user_reset_post_non_existant(self):
        '''An authed post to a non-existant user should receive a not found'''
        add_user_permission(self.user, 'change')
        self.client.force_login(self.user)
        response = self.client.post(self.not_found_url)
        assert response.status_code == 404

    def test_user_reset_post(self):
        '''An authed post should reset the user password'''
        add_user_permission(self.user, 'change')
        self.client.force_login(self.user)
        response = self.client.post(self.url)
        assert response.status_code == 200
        self.assertEqual(len(mail.outbox), 1)


from django.conf import settings
from django.contrib.auth.models import User as DjangoUser
from django.http import HttpRequest, HttpResponse, JsonResponse
from django.urls import reverse
from django.utils.module_loading import import_string
from django.views.generic import View

def user_details(user: DjangoUser) -> dict:
    """The details of the user to return on success"""
    extra = {}
    add_extra_path = getattr(settings, "ACCOUNT_API_DETAILS", False)
    if add_extra_path:
        try:
            details_func = import_string(add_extra_path)
            extra = details_func(user)
        except ImportError:
            pass

    return dict(
        id=user.pk,
        name=user.get_full_name(),
        **extra
    )


class AcceptJsonMixin:
    """
    Adds method to detect if JSON was requested in the Accept header
    """

    def json_response_requested(self: View) -> bool:
        """ does the request want JSON content back?"""
        if "HTTP_ACCEPT" in self.request.META:
            return self.request.META["HTTP_ACCEPT"] == "application/json"
        return False


def manifest(request: HttpRequest) -> HttpResponse:
    """Return a json encoded dictionary {name: path} for views offered
    by Django Accounts API

    :param request: the Django http request
    :type request: HttpRequest
    :return: json encoded name: path dictionary
    :rtype: HttpResponse
    """
    return JsonResponse(dict(
        login=reverse("django_accounts_api:login"),
        logout=reverse("django_accounts_api:logout"),
        password_change=reverse("django_accounts_api:password_change"),
        password_reset=reverse("django_accounts_api:password_reset"),
        password_reset_confirm=reverse(
            "django_accounts_api:password_reset_confirm",
            kwargs=dict(uidb64="uidb64", token="token")
        ),

        users=reverse("django_accounts_api:users"),
        user_create=reverse("django_accounts_api:user_create"),
        user_update=reverse(
            "django_accounts_api:user_update",
            kwargs=dict(pk=00000)
        ),
        user_delete=reverse(
            "django_accounts_api:user_delete",
            kwargs=dict(pk=00000)
        ),
        user_reset=reverse(
            "django_accounts_api:user_reset",
            kwargs=dict(pk=00000)
        )
    ))



from django.contrib.auth import get_user_model
from django.contrib.auth.forms import UsernameField
from django.forms import ModelForm

User = get_user_model()

class UnusablePasswordUserCreationForm(ModelForm):
    """
    A form that creates a user, with no privileges, from the given username
    Sets an unusable password, and returns a reset password link
    """

    class Meta:
        model = User
        fields = ("username", "first_name", "last_name", "email")
        field_classes = {"username": UsernameField}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self._meta.model.USERNAME_FIELD in self.fields:
            self.fields[self._meta.model.USERNAME_FIELD].widget.attrs[
                "autofocus"
            ] = True

    def save(
        self,
        commit=True):

        user = super().save(commit=False)
        user.set_password(None)
        if commit:
            user.save()
            if hasattr(self, "save_m2m"):
                self.save_m2m()
        return user


from django.urls import path

from .views.UserReset import UserReset

from .views.UserCreate import UserCreate
from .views.base import manifest
from .views.login import (
    Login,
    Logout,
    login_check,
)
from .views.password import (
    APIPasswordChange,
    APIPasswordResetView,
    APIPasswordResetConfirmView,
)
from .views.users import (
    APIUsersView, 
    UserUpdate,
    UserDelete
)

app_name = 'django_accounts_api'
urlpatterns = [
    path('', manifest, name='manifest'),

    path('check', login_check, name='login_check'),
    path('login', Login.as_view(), name='login'),
    path('logout', Logout.as_view(), name='logout'),

    path('password_change', APIPasswordChange.as_view(), name='password_change'),

    path('password_reset', APIPasswordResetView.as_view(), name='password_reset'),
    path('reset/<uidb64>/<token>/', APIPasswordResetConfirmView.as_view(), name='password_reset_confirm'),

    path('users/', APIUsersView.as_view(), name='users'),
    path('user/<int:pk>/update', UserUpdate.as_view(), name='user_update'),
    path('user/<int:pk>/delete', UserDelete.as_view(), name='user_delete'),
    path('user/create', UserCreate.as_view(), name='user_create'),
    path('user/<int:pk>/reset', UserReset.as_view(), name='user_reset'),
]



from typing import Any

from django.conf import settings
from django.contrib.auth import login as auth_login
from django.contrib.auth.views import (
    PasswordChangeView, PasswordResetView,
    PasswordResetConfirmView,
    INTERNAL_RESET_SESSION_TOKEN
)
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest, HttpResponse, JsonResponse, HttpResponseRedirect
from django.utils.decorators import method_decorator
from django.urls.exceptions import NoReverseMatch
from django.views.decorators.debug import sensitive_post_parameters
from django.views.decorators.csrf import csrf_protect

from .base import AcceptJsonMixin, user_details


@method_decorator(sensitive_post_parameters(), name='dispatch')
@method_decorator(csrf_protect, name='dispatch')
class APIPasswordChange(AcceptJsonMixin, PasswordChangeView):
    ''' Override the Django change password view to support API use

    GET:

    - default: renders a partial change password form
    - json requested: TODO - Not Implemented yet - json form schema?

    POST - success: 200, failure: 400

    - default: renders a password form with errors
    - json requested: returns password form errors
    '''
    template_name = "django_accounts_api/password_change.html"

    def dispatch(self, request: HttpRequest, *args, **kwargs):
        """Django's PasswordChangeView is login required and redirects, we suppress this and 401"""
        if not request.user.is_authenticated:
            return HttpResponse(status=401)
        return super().dispatch(request, *args, **kwargs)

    def get(self, request: HttpRequest, *args: str, **kwargs: Any) -> HttpResponse:
        """Override the get behavior if json requested TODO json schema"""
        if request.user.is_authenticated:
            if (self.json_response_requested()):
                raise NotImplementedError()
            else:
                return super().post(request, *args, **kwargs)

    def form_invalid(self, form):
        """Override redirect behavior if json is requested return json errors"""
        if self.json_response_requested():
            return JsonResponse(dict(errors=form.errors), status=400)
        else:
            return super().form_invalid(form)

    def form_valid(self, form):
        """Override redirect behavior just return 200 OK"""
        try:
            _repressed_redirect = super().form_valid(form)  # noqa: F841
        except NoReverseMatch:
            pass
        return HttpResponse(status=200)


@method_decorator(sensitive_post_parameters(), name='dispatch')
@method_decorator(csrf_protect, name='dispatch')
class APIPasswordResetView(AcceptJsonMixin, PasswordResetView):
    ''' Override the Django password reset view to support API use

    GET:

    - default: renders a partial password reset form
    - json requested: return a json form schema

    POST - success: 200, failure: 400

    - default: renders a partial password reset form with errors
    - json requested: returns password reset form errors
    '''

    template_name = "django_accounts_api/password_reset.html"

    def get(self, request, *args, **kwargs):
        """Override the get behavior if json requested return json schema"""
        if self.json_response_requested():
            return JsonResponse({
                'schema': [
                    dict(
                        type="email",
                        name="email",
                        label="Email address",
                        help="Please enter your email address.",
                        validation="required|email",
                        placeholder="enter your email",
                    )
                ]})
        else:
            return super().get(request, *args, **kwargs)

    def form_invalid(self, form):
        """Override redirect behavior to return 400
        if json is requested return json errors
        """
        if self.json_response_requested():
            return JsonResponse(dict(errors=form.errors), status=400)
        else:
            response = super().form_invalid(form)
            response.status_code = 400
            return response

    def form_valid(self, form):
        """Override form_valid method to return 200 OK"""
        try:
            _repressed_redirect = super().form_valid(form)  # noqa: F841
        except NoReverseMatch:
            pass
        return HttpResponse(status=200)


@method_decorator(csrf_protect, name='dispatch')
class APIPasswordResetConfirmView(AcceptJsonMixin, PasswordResetConfirmView):
    ''' Override the Django password reset confirm view to support API use

    GET:

    - validates and stores token in session just like django base view
    - validation success redirects to a configured non-django URL default '/reset-password/'
    - validation failure redirects to a configurable non-django URL default '/reset-password/invalid/'

    POST - success: 200, failure: 400

    - default: renders a partial password reset confirm form with errors
    - json requested: returns password reset confirm form errors
    '''
    def __init__(self, **kwargs: Any) -> None:
        # We read password change post login settings to control whether the user is logged in
        # on successful password change
        if 'post_reset_login' not in kwargs:
            kwargs['post_reset_login'] = getattr(settings, 'CHANGE_PASSWORD_POST_LOGIN', False)
        if 'post_reset_login_backend' not in kwargs:
            kwargs['post_reset_login_backend'] = getattr(settings, 'CHANGE_PASSWORD_POST_LOGIN_BACKEND', None)
        # We read the intended frontend reset path from the settings
        self.accounts_api_reset_path = getattr(settings, 'ACCOUNTS_API_RESET_FRONTEND_PATH', '/reset-password/')
        self.accounts_api_reset_fail_path = getattr(settings, 'ACCOUNTS_API_RESET_FAIL_FRONTEND_PATH', '/reset-password/')

        super().__init__(**kwargs)

    template_name = 'django_accounts_api/password_reset_confirm.html'

    def dispatch(self, *args: Any, **kwargs: Any) -> HttpResponse:
        ''' Code duplicated from django/contrib/auth/views.PasswordResetConfirmView.dispatch
        We need to change this to redirect to the frontend ( where marked - Django Accounts API Change )
        '''
        if "uidb64" not in kwargs or "token" not in kwargs:
            raise ImproperlyConfigured(
                "The URL path must contain 'uidb64' and 'token' parameters."
            )

        self.validlink = False
        self.user = self.get_user(kwargs["uidb64"])

        if self.user is not None:
            token = kwargs["token"]
            if token == self.reset_url_token:
                session_token = self.request.session.get(INTERNAL_RESET_SESSION_TOKEN)
                if self.token_generator.check_token(self.user, session_token):
                    # If the token is valid, display the password reset form.
                    self.validlink = True
                    return super().dispatch(*args, **kwargs)
            else:
                if self.token_generator.check_token(self.user, token):
                    # Store the token in the session and redirect to the
                    # password reset form at a URL without the token. That
                    # avoids the possibility of leaking the token in the
                    # HTTP Referer header.
                    self.request.session[INTERNAL_RESET_SESSION_TOKEN] = token
                    # Django Accounts API Change Start
                    # Redirects the user to the frontend for the password change form
                    redirect_url = f'{self.accounts_api_reset_path}{kwargs["uidb64"]}/{self.reset_url_token}'                    #
                    # Django Accounts API Change End
                    return HttpResponseRedirect(redirect_url)

        # Django Accounts API Change Start
        return self.invalid_code()
        # Django Accounts API Change End

    def invalid_code(self):
        ''' Redirect to the frontend invliad code path
        '''
        redirect_url = f'{self.accounts_api_reset_path}invalid/'                    #
        return HttpResponseRedirect(redirect_url)

    def form_valid(self, form):
        """
        If the form is valid, duplicate django's code, avoiding
        reversing the password_reset_complete url which may not exist
        """
        user = form.save()
        del self.request.session[INTERNAL_RESET_SESSION_TOKEN]
        if self.post_reset_login:
            auth_login(self.request, user, self.post_reset_login_backend)

        response =  HttpResponse(status=200)
        response.reason_phrase='password reset confirmed'
        return response

    def form_invalid(self, form):
        """
        If the form is invalid, re-render the context data with the
        data-filled form and errors.
        """
        if self.json_response_requested():
            response = JsonResponse({'errors': form.errors}, status=400)
        else:
            response = super().form_invalid(form)
            response.status_code = 400
        return response



from typing import Any
from django_accounts_api.views.users import User, _user_has_user_perm


from django.contrib.auth.tokens import default_token_generator
from django.contrib.sites.shortcuts import get_current_site
from django.core.mail import EmailMultiAlternatives
from django.http import HttpRequest, HttpResponse, HttpResponseForbidden
from django.views import View
from django.views.generic.detail import SingleObjectMixin
from django.urls import reverse
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode
from django.template import loader


class ResetPasswordMixin():
    subject_template_name="registration/password_reset_subject.txt"
    email_template_name="django_accounts_api/set_password_email.html"
    html_email_template_name="django_accounts_api/set_password_email_html.html"
    from_email=None
    extra_email_context=None
    
    def set_reset_link(self, user):
        # Generate a one-use only link for resetting password and send it to the user.
        use_https = self.request.is_secure(),
        token_generator=default_token_generator
        current_site = get_current_site(self.request)
        domain = current_site.domain
        uid = urlsafe_base64_encode(force_bytes(user.pk))
        token = token_generator.make_token(user)
        reset_path = reverse('django_accounts_api:password_reset_confirm', kwargs=dict(
            uidb64=uid,
            token=token
        ))
        protocol = "https" if use_https else "http"
        self.reset_link = f"{protocol}://{domain}{reset_path}"

        site_name = current_site.name
        email_field_name = User.get_email_field_name()
        user_email = getattr(user, email_field_name)
        if user_email:
            context = {
                "email": user_email,
                "site_name": site_name,
                "uid": uid,
                "user": user,
                "reset_link": self.reset_link,
                **(self.extra_email_context or {}),
            }
            self.sent_email = self.send_mail(
                self.subject_template_name,
                self.email_template_name,
                context,
                self.from_email,
                user_email,
                html_email_template_name=self.html_email_template_name,
            )

    def send_mail(
        self,
        subject_template_name,
        email_template_name,
        context,
        from_email,
        to_email,
        html_email_template_name=None,
    ):
        """
        Send a django.core.mail.EmailMultiAlternatives to `to_email`.
        """
        subject = loader.render_to_string(subject_template_name, context)
        # Email subject *must not* contain newlines
        subject = "".join(subject.splitlines())
        body = loader.render_to_string(email_template_name, context)

        email_message = EmailMultiAlternatives(subject, body, from_email, [to_email])
        if html_email_template_name is not None:
            html_email = loader.render_to_string(html_email_template_name, context)
            email_message.attach_alternative(html_email, "text/html")

        return email_message.send()

class UserReset(SingleObjectMixin, ResetPasswordMixin, View):
    model = User
    def dispatch(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        if not request.user.is_authenticated:
            return HttpResponse(status=401)
        if not _user_has_user_perm(request.user, 'change'):
            return HttpResponseForbidden()
        return super().dispatch(request, *args, **kwargs)

    def post(self,request, *args, **kwargs):
        self.user = self.get_object()
        self.set_reset_link(self.user)
        return HttpResponse(self.reset_link, status=200)
